; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.

; Additional port types

; Keeping track of a port's current row and column.

(define-record-type port-location
  (port          ; the actual port
   )
  ((row 0)       ; current row and column
   (column 0)))

(define make-port-location port-location-maker)

(define (update-row-and-column location char)
  (cond ((eof-object? char) (values))
	((char=? char #\newline)
	 (set-port-location-row! location (+ 1 (port-location-row location)))
	 (set-port-location-column! location 0))
	(else
	 (set-port-location-column! location
				    (+ 1 (port-location-column location))))))

(define (update-row-and-column-from-thing location thing start count)
  (let ((access (if (string? thing)
		    string-ref
		    (lambda (bv i)
		      (ascii->char (code-vector-ref thing i))))))
    (let loop ((i start)
	       (row (port-location-row location))
	       (column (port-location-column location)))
      (cond ((>= i (+ start count))
	     (set-port-location-row!    location row)
	     (set-port-location-column! location column))
	    ((char=? #\newline (access thing i))
	     (loop (+ i 1) (+ row 1) 0))
	    (else
	     (loop (+ i 1) row (+ column 1)))))))

(define (current-row port)
  (if (port-location? (port-data port))
      (port-location-row (port-data port))
      #f))

(define (current-column port)
  (if (port-location? (port-data port))
      (port-location-column (port-data port))
      #f))

;----------------------------------------------------------------
; Input ports that keep track of the current row and column.

; These are not buffered.  A perhaps faster but definitely more complex method
; would be to buffer the input and calculate the row and column only when it
; was requested.

(define tracking-input-port-handler
  (let ((real-port (lambda (port)
		     (port-location-port (port-data port)))))
    (make-input-port-handler
     (lambda (port)
       (list 'tracking-port (real-port port)))
     (lambda (port)
       (make-input-port-closed! port)
       (close-input-port (real-port port)))
     (lambda (port)
       (let ((char (read-char (real-port port))))
	 (update-row-and-column (port-data port) char)
	 char))
     (lambda (port)
       (peek-char (real-port port)))
     (lambda (port)
       (char-ready? (real-port port)))
     (lambda (thing start count port)
       (let ((got (read-block thing start count (real-port port))))
	 (update-row-and-column-from-thing (port-data port) thing start got)
	 (- count got))))))  ; number we didn't get

(define (make-tracking-input-port port)
  (make-port tracking-input-port-handler
	     open-input-port-status
	     (make-port-location port)
	     (make-code-vector 0 0)    ; unbuffered
	     0
	     0
	     #f #f))  ; output stuff

;----------------------------------------------------------------
; Output ports that keep track of the current row and column.

; As with the input ports these are not buffered but could be.

(define tracking-output-port-handler
  (let ((real-port (lambda (port)
		     (port-location-port (port-data port)))))
    (make-output-port-handler
     (lambda (port)
       (list 'tracking-port (real-port port)))
     (lambda (port)
       (close-output-port (real-port port))
       (make-output-port-closed! port))
     (lambda (char port)
       (write-char char (real-port port))
       (update-row-and-column (port-data port) char))
     (lambda (thing start count port)
       (write-block thing start count (real-port port))
       (update-row-and-column-from-thing (port-data port) thing start count))
     (lambda (port)
       (force-output (real-port port))))))

(define (make-tracking-output-port port)
  (make-port tracking-output-port-handler
	     open-output-port-status
	     (make-port-location port)
	     #f #f #f ; input stuff
	     (make-code-vector 0 0)    ; unbuffered
	     0))

(define (fresh-line port)
  (if (and (port-location? (port-data port))
	   (< 0 (port-location-column (port-data port))))
      (newline port)))

;----------------------------------------------------------------
; String input ports

; All the work is done by the VM.  The read methods don't get called until
; the string is used up.

(define string-input-port-handler
  (make-input-port-handler
   (lambda (port)
     (list 'string-input-port))
   (lambda (port)
     (make-input-port-closed! port))
   (lambda (port)
     (eof-object))
   (lambda (port)
     (eof-object))
   (lambda (port)
     #t)
   (lambda (thing start count port)
     count)))

(define (make-string-input-port string)
  (let ((buffer (make-code-vector (string-length string) 0)))
    (copy-chars->bytes! string 0 buffer 0 (string-length string))
    (make-port string-input-port-handler
	       open-input-port-status
	       #f                       ; no additional state needed
	       buffer
	       0
	       (string-length string)   ; number of bytes available
	       #f #f)))  ; output stuff

;----------------------------------------------------------------
; String output ports

; The data field of the port is a list of full buffers.  When the
; port is closed these are concatenated together to get the final
; string.

(define buffer-size 1024)

; Concatenates all of the buffers into single string.

(define (string-output-port-output port)
  (let* ((full (port-data port))
	 (last (port-out-buffer port))
	 (index (port-out-index port))
	 (count (apply + index (map code-vector-length full)))
	 (out (make-string count)))
    (let loop ((full (reverse full)) (i 0))
      (if (null? full)
	  (copy-bytes->chars! last 0 out i index)
	  (let* ((buffer (car full))
		 (count (code-vector-length buffer)))
	    (copy-bytes->chars! buffer 0 out i count)
	    (loop (cdr full) (+ i count)))))
    out))

(define string-output-port-handler
  (make-output-port-handler
   (lambda (port)
     '(string-output-port))
   (lambda (port)
     (make-output-port-closed! port))
   (lambda (char port)
     (set-port-data! port (cons (port-out-buffer port) (port-data port)))
     (set-port-out-buffer! port (make-code-vector buffer-size 0))
     (set-port-out-index! port 0)
     (write-char char port))
   (lambda (thing start count port)
     (let* ((i (port-out-index port))
	    (b (port-out-buffer port))
	    (done (- (code-vector-length b) i))
	    (count (- count done))
	    (new (make-code-vector (+ buffer-size count) 0)))
       (if (string? thing)
	   (copy-chars->bytes! thing start b i done)
	   (copy-bytes! thing start b i done))
       (set-port-data! port (cons b (port-data port)))
       (if (string? thing)
	   (copy-chars->bytes! thing (+ start done) new 0 count)
	   (copy-bytes! thing (+ start done) new 0 count))
       (set-port-out-buffer! port new)
       (set-port-out-index! port count)))
   (lambda (port)
     (values))))

(define (make-string-output-port)
  (make-port string-output-port-handler
	     open-output-port-status
	     '()
	     #f #f #f ; input stuff
	     (make-code-vector buffer-size 0)
	     0))

(define (call-with-string-output-port proc)
  (let ((port (make-string-output-port)))
    (proc port)
    (string-output-port-output port)))

;----------------------------------------------------------------
; Copying utilities

(define (copy-bytes! from i to j count)
  (let ((limit (+ count i)))
    (do ((i i (+ i 1))
	 (j j (+ j 1)))
	((= i limit))
      (code-vector-set! to j (code-vector-ref from i)))))

(define (copy-chars->bytes! from i to j count)
  (let ((limit (+ count i)))
    (do ((i i (+ i 1))
	 (j j (+ j 1)))
	((= i limit))
      (code-vector-set! to j (char->ascii (string-ref from i))))))

(define (copy-bytes->chars! from i to j count)
  (let ((limit (+ count i)))
    (do ((i i (+ i 1))
	 (j j (+ j 1)))
	((= i limit))
      (string-set! to j (ascii->char (code-vector-ref from i))))))

;----------------------------------------------------------------
; Output ports from a single character writer

(define char-at-a-time-port-handler
  (make-output-port-handler
   (lambda (port)
     (list 'output-port (port-data port)))
   (lambda (port)
     (make-output-port-closed! port))
   (lambda (char port)
     ((port-data port) char))
   (lambda (thing start count port)
     (let ((proc (port-data port))
	   (limit (+ start count)))
       (do ((i start (+ i 1)))
	   ((= i limit))
	 (proc (if (string? thing)
		   (string-ref thing i)
		   (ascii->char (code-vector-ref thing i)))))))
   (lambda (port)
     (values))))

(define (make-char-at-a-time-output-port proc)
  (make-port char-at-a-time-port-handler
	     open-output-port-status
	     proc
	     #f #f #f ; input stuff
	     (make-code-vector 0 0)    ; unbuffered
	     0))

(define (write-one-line port count proc)
  (call-with-current-continuation
    (lambda (quit)
      (proc (make-char-at-a-time-output-port
	     (lambda (char)
	       (write-char char port)
	       (set! count (- count 1))
	       (if (<= count 0)
		   (quit #f))))))))
