; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.

; Multitasking

; A thread has:
;   - dynamic environment
;   - continuation: #f if the thread is current, 'done if it is finished,
;     and a primitive continuation (which may be #F) otherwise.
;   - env-for-spawn - initial dynamic environment for threads spawned
;     by this thread.
;   - cells - a list of queue entries containing pairs (<the-thread> .
;     <return-thunk>).  When the thread is resumed whatever the appropriate
;     <return-thunk> returns is passed to the thread.  When the thread is
;     moved to a new queue all of the old queue entries are delected from
;     their respective queues.
;   - uid - no semantic content.
;   - name - no semantic content.

;   Date: Mon, 31 Aug 92 18:17:23 -0400
;   From: "Guillermo J. Rozas" <jinx@martigny.ai.mit.edu>
;
;   1: Instead of requesting a periodic timer interrupt from the system,
;   request one-shot interrupts, and re-request at the end of each throw
;   (in the argument to within-continuation).  Thus the time to "task
;   switch" becomes "system time" and each thread gets the full time
;   slice.
;
; The cost of the above Rozas hack is one system call per time slice.
; This turns out to be a significant percentage of the time it takes
; to swap threads.
;
; What we do instead is schedule a periodic interrupt at five times the
; requested thread quantum and then switch every fifth tick.

; Threads

(define-record-type thread :thread
  (really-make-thread dynamic-env
		      continuation env-for-spawn
		      cells enabled-ints uid name)
  thread?
  (dynamic-env   thread-dynamic-env)  ;Must be first!  (See fluid.scm)
  (continuation  thread-continuation set-thread-continuation!)
  (env-for-spawn thread-env-for-spawn)
  (cells         thread-cells set-thread-cells!)
  (enabled-ints  thread-enabled-ints set-thread-enabled-ints!)
  (uid           thread-uid)   ;for debugging
  (name		 thread-name)) ;for debugging

(define *thread-uid* 0)

(define (make-thread thunk env-for-spawn name)
  (let ((thread (really-make-thread env-for-spawn
				    (compose-continuation
				     (lambda (ignore)
				       (thread-top-level thunk))
				     #f)
				    env-for-spawn
				    '()
				    all-interrupts
				    *thread-uid*
				    name)))
    (set! *thread-uid* (+ *thread-uid* 1))
    thread))

; CELLS
; A cell contains a thread and a thunk.  When the thread is resumed it is
; passed the value(s) returned by the thunk.  When a thread is moved from
; one set of queues etc. to another its old cells are invalidated by replacing
; the thread with #F.

(define (add-to-queue thread thunk queue)
  (let ((cell (cons thread thunk)))
    (set-thread-cells! thread (cons (enqueue queue cell) (thread-cells thread)))
    cell))

(define cell-thread car)
(define cell-thunk cdr)

; Optmized for the very common one element case.
; This need not be called on the current thread, as that is never on
; any queues.

(define (detach-thread! thread)
  (let ((cells (thread-cells thread)))
    (if (not (null? cells))                   ;++ makes a significant speedup
	(begin
	  (delete-queue-entry! (car cells))
	  (if (not (null? (cdr cells)))       ;++ makes a significant speedup
	      (for-each (lambda (cell)
			  (delete-queue-entry! cell))
			(cdr cells))))))
  (set-thread-cells! thread '()))

; A thunk for when you don't care what value a thread gets.

(define unspecific
  (let ((x (if #f #f)))
    (lambda () x)))

;------------

(define (enter-kernel!)
  (set-thread-enabled-ints! (current-thread)
			    (set-enabled-interrupts! no-interrupts)))

(define (exit-kernel!)
  (set-enabled-interrupts! (thread-enabled-ints (current-thread))))

(define (thread-top-level thunk)
  (thunk)
  (finish-current-thread))

(define (finish-current-thread)
  (enter-kernel!)
  (switch-to-another-thread #f))

; Spawn a new thread

(define (spawn thunk . name-option)
  (let* ((name (if (null? name-option) #f (car name-option)))
	 (thread (make-thread thunk
			      (thread-env-for-spawn (current-thread))
			      name)))
    (start-thread thread)
    thread))

; Same, but doesn't start the thread immediately (useful if the
; current thread doesn't want to go on the runnable queue).

(define (delayed-spawn thunk . name-option)    ;inside kernel
  (let* ((name (if (null? name-option) #f (car name-option)))
	 (thread (make-thread thunk
			      (thread-env-for-spawn (current-thread))
			      name)))
    (make-ready! thread unspecific)
    thread))

; --------------------
; Scheduler

(define *runnable-threads* (make-queue))

(define (make-ready! thread values-thunk)  ; interrupts disabled
  (detach-thread! thread) 
  (add-to-queue thread values-thunk *runnable-threads*))

(define (make-cell-ready! cell)            ; interrupts disabled
  (make-ready! (cell-thread cell) (cell-thunk cell)))

; Current thread is known not to be on any queues (or we wouldn't be making
; it ready), so we don't have to call DETACH-THREAD!.

(define (make-current-thread-ready!)
  (add-to-queue (current-thread) unspecific *runnable-threads*))

(define *ticks-per-turn* 5)

(define *time-left* *ticks-per-turn*)
(define *quantum-mantissa* 1)    ; 0.1 seconds = 1 * 10^-1
(define *quantum-exponent* -1)

(define (start-periodic-interrupts!)
  (schedule-interrupt *quantum-mantissa* *quantum-exponent* #t))

(define (stop-periodic-interrupts!)
  (schedule-interrupt 0 0 #f))

(define (handle-timer-interrupt template ei)
  ;; Ignore the template, we don't care.
  ;; Interrupts are disabled at this point, they will be enabled automatically
  ;; when this procedure returns.
  (if (= *time-left* 0)
      (dispatch)
      (set! *time-left* (- *time-left* 1))))

(define (dispatch)				;Call with interrupts disabled
  (make-current-thread-ready!)
  (suspend-this-thread))

(define (relinquish-timeslice)
  (enter-kernel!)
  (dispatch))

(define (suspend-this-thread)			;Call with interrupts disabled
  (switch-to-another-thread #t))

; Pull the next thread off of the runnable queue and run it instead of the
; current thread.  If RESUME? is #F the current thread is finished and the
; current continuation need not be saved.

(define (switch-to-another-thread resume?)	;Call with interrupts disabled
  (if (queue-empty? *runnable-threads*)
      (begin (set-enabled-interrupts! all-interrupts)
	     (signal 'deadlock)
	     (error "no threads to run"))
      (let ((next (dequeue *runnable-threads*)))
	(cond (resume?
	       (switch-to-thread (cell-thread next) (cell-thunk next)))
	      (else
	       ;; #F is a legit continuation, so DONE is used instead
	       (set-thread-continuation! (current-thread) 'done)
	       (schedule-thread (cell-thread next) (cell-thunk next)))))))

(define (switch-to-thread thread values-thunk)	;Call with interrupts disabled
  (primitive-cwcc
   (lambda (cont)
     (set-thread-continuation! (current-thread) cont)
     (schedule-thread thread values-thunk))))

(define (schedule-thread thread values-thunk)	;Call with interrupts disabled
  (set-thread-cells! thread '())    ; known not to be on any queues
  (set-current-thread! thread)
  (set! *time-left* *ticks-per-turn*)
  (let ((cont (thread-continuation (current-thread))))
    (set-thread-continuation! (current-thread) #f)
    (set-enabled-interrupts! (thread-enabled-ints thread))
    (with-continuation cont values-thunk)))

; --------------------
; Initialization

; (with-multitasking thunk) initializes for multitasking, then starts
; up a thread for the execution of THUNK.  That thread and all others
; created will run in the dynamic context of the call to
; with-multitasking.  The call to with-multitasking finally returns
; only when the scheduler runs out of things to do.

(define (with-multitasking thunk)
  (set! *runnable-threads* (make-queue))
  (initialize-sleepers) ; drop any leftover sleeping threads
  (with-interrupts-inhibited
   (lambda ()
     (let ((swap! (let ((save-blocked-handler handle-blocked-channel))
		    (lambda ()
		      (set! save-blocked-handler
			    (set-blocking-channel-handler!
			     save-blocked-handler))))))
       (dynamic-wind
	(lambda ()
	  (swap!)
	  (vector-set! interrupt-handlers
		       (enum interrupt alarm)
		       handle-timer-interrupt)
	  (start-periodic-interrupts!))
	(lambda ()
	  (start-multitasking thunk))
	(lambda ()
	  (stop-periodic-interrupts!)
	  (reset-timer-interrupts!)
	  (swap!)))))))

; Add handlers for deadlock and termination.

(define (start-multitasking thunk)
  (call-with-current-continuation
    (lambda (deadlock)
      (with-handler
       (lambda (c punt)
	 (if (deadlock? c) (deadlock 0) (punt)))
       (lambda ()
	 (call-with-current-continuation
	   (lambda (terminate)
	     (with-handler (lambda (c punt)
			     (if (terminate? c)
				 (terminate 0)
				 (punt)))
			   (lambda ()
			     (make-ready!
			      (make-thread wakeup-service (get-dynamic-env) 'wakeup-service)
			      unspecific)
			     ;; Schedule-thread never returns.
			     (schedule-thread
			      (make-thread thunk (get-dynamic-env) 'initial)
			      unspecific)))))
	 ;; Interrupts enabled here, probably
	 (finish-current-thread))))))

(define-condition-type 'terminate '())
(define terminate? (condition-predicate 'terminate))
(define-condition-type 'deadlock '())
(define deadlock? (condition-predicate 'deadlock))

(define no-interrupts 0)

; --------------------
; Random operations: start, stop, kill, interrupt
; STOP is dangerous because it removes the thread from wherever it currently
; resides, such as waiting for a lock.  STARTing it up again will allow
; it to run while erroneously thinking it has the lock.

(define (stopped? thread)
  (and (null? (thread-cells thread))
       (not (eq? thread (current-thread)))))

(define (done? thread)
  (eq? (thread-continuation thread) 'done))

(define (active? thread)
  (not (or (done? thread)
	   (stopped? thread))))

(define (terminate-current-thread)
  (signal 'terminate)
  (error "can't terminate current thread")
  0)    ;suppress compiler warning

(define (start-thread thread)
  (if (thread? thread)
      (begin (enter-kernel!)
	     (cond ((stopped? thread)
		    (make-current-thread-ready!)
		    (switch-to-thread thread unspecific)
		    #t)
		   (else
		    (exit-kernel!)
		    (active? thread))))
      (call-error "invalid argument" start-thread thread)))

; Dangerous!!

(define (stop-thread thread)
  (if (thread? thread)
      (begin (enter-kernel!)
	     (detach-thread! thread)
	     (cond ((active? thread)
		    (if (eq? thread (current-thread))
			(suspend-this-thread))
		    #t)
		   (else
		    (exit-kernel!)
		    (stopped? thread))))
      (call-error "invalid argument" stop-thread thread)))

(define (kill-thread thread)
  (interrupt-thread thread terminate-current-thread))

(define (interrupt-thread thread thunk)
  (if (thread? thread)
      (with-interrupts-inhibited
	(lambda ()
	  (cond ((eq? thread (current-thread))
		 (thunk))
		((stop-thread thread)
		 (set-thread-continuation!
		  thread
		  (compose-continuation
		   (lambda (ignore)
		     (thunk))
		   (or (thread-continuation thread)
		       (error "this shouldn't happen" thread))))
		 (start-thread thread))
		(else
		 #f))))
      (call-error "invalid argument" interrupt-thread thread)))

; --------------------
; Events
; An event is something a thread can wait for.  These are really event types,
; for example locks.  Waiting for one requires giving an instance of the thing
; itself, such as a lock.  These consist of three procedures:
;  (INITIALIZE instance) - called before waiting; if this returns #F the
;     thread needn't wait at all.
;  (REGISTER cell instance) - register the cell as waiting for the instance
;     of the event.
;  (VALUE instance) - returns the value(s) of the instance once the thread
;     has finished waiting (or has been told that it needn't wait).

(define-record-type event :event
  (make-event type initialize register value)
  event?
  (type event-type)
  (initialize event-initialize)
  (register   event-register)
  (value      event-value))

; Waiting for an event:
;  1. Disable interrupts
;  2. Initialize the event - if waiting is necessary:
;      a. detach the current thread (from wherever it may be)
;      b. register it as waiting
;      c. suspend
;  3. Enable interrupts

(define (wait-for-event event data after)
  (if (event? event)
      (begin (enter-kernel!)
	     (cond (((event-initialize event) (current-thread) data)
		    ((event-register event) (current-thread) after data)
		    (suspend-this-thread))
		   (else
		    ((event-value event) data after))))
      (call-error "invalid argument" wait-for-event event)))

; STUFF is a list of events and instances.  This waits until the first
; one finishes.

;(define (wait-for-events . stuff)
;  (if (null? stuff)
;      (values)
;      (let ((ei (set-enabled-interrupts! no-interrupts)))
;        (let loop ((stuff stuff))
;          (cond ((null? stuff)
;                 (really-wait-for-events stuff ei))
;                ((or (null? (cdr stuff))
;                     (not (event? (car stuff))))
;                 (set-enabled-interrupts! ei)
;                 (call-error "invalid arguments" wait-for-events stuff))
;                (((event-initialize (car stuff)) (current-thread) (cadr stuff))
;                 (loop (cddr stuff)))
;                (else
;                 (set-enabled-interrupts! ei)
;                 ((event-value (car stuff)) (cadr stuff))))))))
;
;(define (really-wait-for-events stuff ei)
;  (detach-thread! (current-thread))
;  (do ((i 0 (+ i 1))
;       (stuff stuff (cddr stuff)))
;      ((null? stuff))
;    ((event-register (car stuff))
;     (make-cell (current-thread) (lambda () i))
;     (cadr stuff)))
;  (let ((hit (suspend-this-thread)))
;    (set-enabled-interrupts! ei)
;    (do ((i 0 (+ i 1))
;         (stuff stuff (cddr stuff)))
;        ((= hit i)
;         ((event-value (car stuff)) (cadr stuff))))))
		   
; --------------------
; Locks (= semaphores)

(define-record-type lock :lock
  (really-make-lock owner queue uid)
  lock?
  (owner lock-owner set-lock-owner!)
  (queue lock-queue)
  (uid lock-uid))

(define *lock-uid* 0)

(define (make-lock)
  (let ((uid *lock-uid*))
    (set! *lock-uid* (+ uid 1))
    (really-make-lock #f (make-queue) uid)))

(define (lock-initialize thread lock)	; called with interrupts disabled
  (if (not (lock? lock))
      (begin (exit-kernel!)
	     (error "not a lock" lock))
      (let ((owner (lock-owner lock)))
	(cond ((not owner)
	       (set-lock-owner! lock (current-thread))
	       #f)
	      ((eq? (current-thread) (lock-owner lock))
	       (exit-kernel!)
	       (error "attempting to obtain an already owned lock" lock))
	      ((done? owner)
	       ;; not good - interrupts are disabled
	       (warn "taking lock from terminated owner" lock)
	       (set-lock-owner! lock (current-thread))
	       #f)
	      (else
	       #t)))))  ; have to wait

(define (lock-register thread thunk lock)
  (add-to-queue thread thunk (lock-queue lock)))

(define (lock-value lock after)
  (exit-kernel!)
  (after))

(define lock-event (make-event 'lock lock-initialize lock-register lock-value))

;(define (obtain-lock lock)
;  (wait-for-event lock-event lock unspecific))

; Hand hacked for a bit of a speed increase.  Am I a loser or what?
; This is enough to make a lock-intensive benchmark run a smidgen
; faster in the new code than in the old.

(define (obtain-lock lock)
  (enter-kernel!)
  (cond ((lock-initialize (current-thread) lock)
	 (lock-register (current-thread) unspecific lock)
	 (suspend-this-thread))
	(else
	 (lock-value lock unspecific))))

; WITH-INTERRUPTS-INHIBITED has been hand integrated here.  It makes a
; difference.

(define (release-lock lock)		;Interrupts enabled
  (if (eq? (lock-owner lock) (current-thread))
      (begin (enter-kernel!)
	     (set-lock-owner! lock #f)
	     (release-locked-thread lock (lock-queue lock)))
      (error "lock owned by another thread" lock)))

(define (release-locked-thread lock queue)    ;Interrupts disabled
  (if (queue-empty? queue)
      (exit-kernel!)
      (let ((next (dequeue queue)))
	(set-lock-owner! lock (cell-thread next))
	(make-current-thread-ready!)
	(switch-to-thread (cell-thread next)
			  (cell-thunk next)))))

(define (with-lock lock thunk)		;Interrupts enabled
  (if (eq? (lock-owner lock) (current-thread))
      (thunk)
      (dynamic-wind (lambda () (obtain-lock lock))
		    thunk
		    (lambda () (release-lock lock)))))


; --------------------
; Condition variables (single-assignment cells)

(define-record-type condvar :condvar
  (really-make-condvar queue id)
  condvar?
  (queue condvar-queue set-condvar-queue!)  ; #f means it has been set
  (value condvar-value set-condvar-value!)
  (id condvar-id))

(define (make-condvar . id-option)
  (really-make-condvar (make-queue)
		       (if (null? id-option) #f (car id-option))))

(define (condvar-initialize thread condvar)  ; called with interrupts disabled
  (cond ((not (condvar? condvar))
	 (set-enabled-interrupts! all-interrupts)
	 (error "not a condvar" condvar))
	((condvar-queue condvar)
	 #t) ; have to wait
	(else
	 #f)))

(define (condvar-register thread thunk condvar)
  (add-to-queue thread thunk (condvar-queue condvar)))

(define (get-condvar-value condvar after)
  (exit-kernel!)
  (after))

(define condvar-event
  (make-event 'condvar condvar-initialize condvar-register get-condvar-value))

(define (condvar-ref condvar)
  (wait-for-event condvar-event condvar (lambda () (condvar-value condvar))))

(define (condvar-set! condvar value)	;Interrupts enabled
  (cond ((condvar? condvar)
	 (enter-kernel!)
	 (let ((q (condvar-queue condvar)))
	   (cond (q
		  (set-condvar-value! condvar value)
		  (set-condvar-queue! condvar #f)
		  (release-condvar-waiters q value))
		 (else
		  (exit-kernel!)
	       ;; OK to store the same value twice.
		  (if (not (eq? value (condvar-value condvar)))
		      (error "condition variable is already assigned"
			     (condvar-value condvar) value))))))
	(else
	 (call-error "invalid argument" condvar-set! condvar value))))

(define (release-condvar-waiters q value)     ;Interrupts disabled
  (if (queue-empty? q)
      (exit-kernel!)
      (let ((first (dequeue q)))
	(do ()
	    ((queue-empty? q))
	  (make-cell-ready! (dequeue q)))
	(make-current-thread-ready!)
	(switch-to-thread (cell-thread first)
			  (lambda () value)))))

(define (kill-condvar condvar)
  (for-each (lambda (cell)
	      (kill-thread (cell-thread cell)))
	    (with-interrupts-inhibited
	     (lambda ()
	       (if (condvar-queue condvar)
		   (queue->list (condvar-queue condvar))
		   '())))))

; --------------------
; Silliness

(define (active-threads)		;debugging utility
  (with-interrupts-inhibited
    (lambda ()
      (map cell-thread (queue->list *runnable-threads*)))))

(define-record-discloser :thread
  (lambda (thread)
    (cons 'thread
	  (cons (thread-uid thread)
		(let ((name (thread-name thread)))
		  (if name
		      (list name)
		      '()))))))

;(put 'with-interrupts-inhibited 'scheme-indent-hook 0)
;(put 'with-handler 'scheme-indent-hook 1)
