; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.


; Interrupts

(define interrupt-handlers
  (make-vector interrupt-count 0))

(do ((i 0 (+ i 1)))
    ((= i interrupt-count))
  (vector-set! interrupt-handlers i
	       (lambda stuff
		 (apply signal (cons 'interrupt (cons i stuff))))))

(define (initialize-interrupts! thunk)
  (set-interrupt-handlers! interrupt-handlers)
  (set-enabled-interrupts! all-interrupts)
  (reset-timer-interrupts!)
  (thunk))

(define no-interrupts 0)

(define all-interrupts
  (- (arithmetic-shift 1 interrupt-count) 1))

(define (with-interrupts-inhibited thunk)
  (with-interrupts no-interrupts thunk))

(define (with-interrupts-allowed thunk)
  (with-interrupts all-interrupts thunk))

(define (with-interrupts interrupts thunk)
  ;; I might consider using dynamic-wind here, but (a) I'm worried
  ;; about the speed of thread switching (which uses this) and (b)
  ;; it's a pretty bad idea to throw in or out of one of these anyhow.
  (let ((ei (set-enabled-interrupts! interrupts)))
    (call-with-values thunk
      (lambda results
	(set-enabled-interrupts! ei)
	(apply values results)))))

(define (enabled-interrupts)		;For debugging
  (let ((e (set-enabled-interrupts! 0)))
    (set-enabled-interrupts! e)
    e))

; Signal an interrupt if an insufficient amount of memory is reclaimed by
; a garbage collection.  The amount defaults to 10% of the heap.

(define *post-gc-procedures* '())

(define (call-after-gc! thunk)
  (if (not (memq thunk *post-gc-procedures*))
      (set! *post-gc-procedures* (cons thunk *post-gc-procedures*))))

(vector-set! interrupt-handlers
	     (enum interrupt post-gc)
	     (lambda (ei)
	       (set-enabled-interrupts! ei)
	       (for-each (lambda (thunk) (thunk)) *post-gc-procedures*)))

(define *required-post-gc-space* 0)

(define (check-recovered-space)
  (if (> *required-post-gc-space*
	 (memory-status (enum memory-status-option available) 0))
      (signal 'warning
	      'heap-space-shortage
	      (enum memory-status-option available)
	      *required-post-gc-space*)))

(call-after-gc! check-recovered-space)

(define (interrupt-before-heap-overflow! . maybe-required-space)
  (set! *required-post-gc-space*
	(if (null? maybe-required-space)
	    (quotient (memory-status (enum memory-status-option heap-size) 0)
		      10)
	    (car maybe-required-space))))

; Flush output ports every 0.3 seconds.
; JAR lobotomization to 0.39 4/11/95 - disabled due to lack of locking.

(define reset-timer-interrupts!
  (let ((handler (vector-ref interrupt-handlers
			     (enum interrupt alarm))))
    (lambda ()
      (vector-set! interrupt-handlers
		   (enum interrupt alarm)
		   handler))))

; Was:
;  (set-timer-interrupt! 3 -1 #t
;                        (lambda (template ei)
;                          (set-enabled-interrupts! ei)
;                          (force-output-ports!)))

(define set-timer-interrupt!
  (let ((mant 0)
	(exp  0)
	(per? #f))
    (lambda (new-mant new-exp new-per? new-handler)
      (with-interrupts-inhibited
       (lambda ()
	 (let ((save-handler (vector-ref interrupt-handlers
					 (enum interrupt alarm))))
	   (vector-set! interrupt-handlers
			(enum interrupt alarm)
			new-handler)
	   (let ((save-mant mant)
		 (save-exp exp)
		 (save-per? per?))
	     (schedule-interrupt new-mant new-exp new-per?)
	     (set! mant new-mant)
	     (set! exp  new-exp)
	     (set! per? new-per?)
	     (lambda ()
	       (with-interrupts-inhibited
		(lambda ()
		  (vector-set! interrupt-handlers
			       (enum interrupt alarm)
			       save-handler)
		  (schedule-interrupt save-mant save-exp save-per?)))))))))))
