; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.

; Ports and port handlers

(define-record-type port-handler :port-handler
  (make-port-handler
   discloser
   close-input-port read-char peek-char char-ready? read-block
   close-output-port write-char write-block force-output)
  port-handler?
  (discloser         port-handler-discloser)
  (close-input-port  port-handler-close-input-port)
  (read-char         port-handler-read-char)
  (peek-char         port-handler-peek-char)
  (char-ready?       port-handler-char-ready?)
  (read-block        port-handler-read-block)
  (close-output-port port-handler-close-output-port)
  (write-char        port-handler-write-char)
  (write-block       port-handler-write-block)
  (force-output      port-handler-force-output))

(define (make-input-port-handler discloser close
				 read-char peek-char char-ready? read-block)
  (make-port-handler discloser close read-char peek-char char-ready? read-block
		     #f #f #f #f))

(define (make-output-port-handler discloser close
				  write-char write-block force-output)
  (make-port-handler discloser #f #f #f #f #f
		     close write-char write-block force-output))
  
; abstracted to avoid a circular structure dependency

(define (initialize-i/o-handlers! define-exception-handler signal-exception)
  (let-syntax ((define-i/o-handler
		 (syntax-rules ()
		   ((define-i/o-handler (?id ?arg ...) ?port ?handler)
		    (define-exception-handler (enum op ?id)
		      (lambda (opcode reason ?arg ...)
			(if (= reason (enum exception buffer-full/empty))
			    ((?handler (port-handler ?port)) ?arg ...)
			    (signal-exception opcode reason ?arg ...))))))))
    (define-i/o-handler (read-char port) port port-handler-read-char)
    (define-i/o-handler (peek-char port) port port-handler-peek-char)
    (define-i/o-handler (char-ready? port) port port-handler-char-ready?)
    (define-i/o-handler (xread-block thing start count port) port
      port-handler-read-block)
    
    (define-i/o-handler (write-char char port) port port-handler-write-char)
    (define-i/o-handler (write-block thing start count port) port
      port-handler-write-block)))

(define (close-input-port port)
  (if (open-input-port? port)
      ((port-handler-close-input-port (port-handler port)) port)
      (call-error "invalid argument" close-input-port port)))

(define (close-output-port port)
  (if (open-output-port? port)
      ((port-handler-close-output-port (port-handler port)) port)
      (call-error "invalid argument" close-output-port port)))

(define (force-output port)
  (if (open-output-port? port)
      ((port-handler-force-output (port-handler port)) port)
      (call-error "invalid argument" force-output port)))

;----------------------------------------------------------------
; Non-blocking I/O stuff.

(define (wait-for-channel channel)
  (if (null? (find-ready-channels #f #f (list channel)))
      (wait-for-channel channel)))

(define *block-on-channel* wait-for-channel)

(define (block-on-channel channel)
  (*block-on-channel* channel))

(define (set-blocking-channel-handler! proc)
  (let ((old *block-on-channel*))
    (set! *block-on-channel* proc)
    old))

;----------------------------------------------------------------
; Input ports

(define input-port-mask
  (arithmetic-shift 1 (enum port-status-options input)))

(define open-input-port-mask
  (arithmetic-shift 1 (enum port-status-options open-for-input)))

(define open-input-port-status
  (bitwise-ior input-port-mask open-input-port-mask))

(define (open-input-port? port)
  (not (= 0 (bitwise-and open-input-port-mask (port-status port)))))

(define (make-input-port-closed! port)
  (set-port-status! port (bitwise-and (port-status port)
				      (bitwise-not open-input-port-mask))))

(define-record-type input-port-data :input-port-data
  (make-input-port-data channel pending-eof?)
  (channel input-channel)
  (pending-eof? pending-eof? set-pending-eof?!))

(define (make-get-char read?)
  (lambda (port)
     (let ((data (port-data port)))
       (cond ((pending-eof? data)
	      (if read?
		  (set-pending-eof?! data #f))
	      (eof-object))
	     (else
	      (cond ((fill-buffer! port (input-channel data))
		     (if read?
			 (read-char port)
			 (peek-char port)))
		    (else
		     (if (not read?)
			 (set-pending-eof?! data #t))
		     (eof-object))))))))

(define input-channel-handler
  (make-input-port-handler
   (lambda (port)
     (list 'input-port (input-channel (port-data port))))
   (lambda (port)
     (close-channel (input-channel (port-data port)))
     (make-input-port-closed! port))
   (make-get-char #t)
   (make-get-char #f)
   (lambda (port)
     (let ((data (port-data port)))
       (if (pending-eof? data)
	   #t
	   (not (null? (find-ready-channels 0 0 (list (input-channel data))))))))
   (lambda (thing start count port)
     (let ((data (port-data port)))
       (cond ((pending-eof? data)
	      (set-pending-eof?! data #f)
	      count)
	     (else
	      (completely-fill-buffer! thing start count (input-channel data))))))))

(define (input-channel->port channel buffer-size)
  (if (>= 0 buffer-size)
      (call-error "invalid buffer size" input-channel->port channel buffer-size)
      (make-port input-channel-handler
		 (bitwise-ior input-port-mask open-input-port-mask)
		 (make-input-port-data channel #f)
		 (make-code-vector buffer-size 0)
		 0
		 0
		 #f #f)))  ; output stuff
  
(define (fill-buffer! port channel)
  (let ((got (channel-read (port-in-buffer port)
			   0
			   (code-vector-length (port-in-buffer port))
			   channel)))
    (cond ((eof-object? got)
	   #f)
	  ((= 0 got)
	   (block-on-channel channel)
	   (fill-buffer! port channel))
	  (else
	   (set-port-in-index! port 0)
	   (set-port-in-limit! port got)
	   #t))))

(define (completely-fill-buffer! thing start count channel)
  (let loop ((start start) (count count))
    (let ((got (channel-read thing start count channel)))
      (cond ((eof-object? got)
	     count)
	    ((= got count)
	     0)
	    (else
	     (if (= 0 got)
		 (block-on-channel channel))
	     (loop (+ start got) (- count got)))))))

;----------------------------------------------------------------
; Output ports

(define output-port-mask
  (arithmetic-shift 1 (enum port-status-options output)))

(define open-output-port-mask
  (arithmetic-shift 1 (enum port-status-options open-for-output)))

(define open-output-port-status
  (bitwise-ior output-port-mask open-output-port-mask))

(define (open-output-port? port)
  (not (= 0 (bitwise-and open-output-port-mask (port-status port)))))

(define (make-output-port-closed! port)
  (set-port-status! port (bitwise-and (port-status port)
				      (bitwise-not open-output-port-mask))))

(define output-channel-handler
  (make-output-port-handler
   (lambda (port)
     (list 'output-port (port-data port)))
   (lambda (port)
     (force-output port)
     (close-channel (port-data port))
     (make-output-port-closed! port))
   (lambda (char port)
     (write-port-buffer! port)
     (write-char char port))
   (lambda (thing start count port)
     (write-port-buffer! port)
     (write-buffer thing start count port))
   (lambda (port)
     (if (< 0 (port-out-index port))
	 (write-port-buffer! port)))))

(define unbuffered-output-channel-handler
  (make-output-port-handler
   (lambda (port)
     (list 'output-port (port-data port)))
   (lambda (port)
     (close-channel (port-data port))
     (make-output-port-closed! port))
   (lambda (char port)
     (let ((buffer (port-in-buffer port)))  ; one-char buffer stashed here
       (code-vector-set! buffer 0 (char->ascii char))
       (write-buffer buffer 0 1 port)))
   (lambda (thing start count port)
     (write-buffer thing start count port))
   (lambda (port)
     (values))))

(define (output-channel->port channel buffer-size)
  (if (> 0 buffer-size)
      (call-error "invalid buffer size"
		  output-channel->port channel buffer-size)
      (let ((port (make-port (if (= 0 buffer-size)
				 unbuffered-output-channel-handler
				 output-channel-handler)
			     (bitwise-ior output-port-mask open-output-port-mask)
			     channel
			     (if (= 0 buffer-size)  ; hide a buffer in the input buffer field
				 (make-code-vector 1 0)
				 #f)
			     #f #f ; input stuff
			     (make-code-vector buffer-size 0)
			     0)))
	(if (< 0 buffer-size)
	    (periodically-force-output! port))
	port)))
	     
(define (write-port-buffer! port)
  (write-buffer (port-out-buffer port) 0 (port-out-index port) port)
  (set-port-out-index! port 0))

(define (write-buffer buffer start count port)
  (let ((channel (port-data port)))
    (let loop ((start start) (count count))
      (let ((sent (channel-write buffer start count channel)))
	(cond ((< sent count)
	       (if (= 0 sent)
		   (block-on-channel channel))
	       (loop (+ start sent) (- count sent))))))))

;----------------------------------------------------------------

; Code to ensure that output ports are flushed periodically.

(define *flush-these-ports* (cons #f '()))

(define (initialize-output-port-list!)
  (set! *flush-these-ports* (cons #f '())))

(define (periodically-force-output! port)
  (set-cdr! *flush-these-ports*
	    (cons (make-weak-pointer port) (cdr *flush-these-ports*))))

; The interrupt code ensures that this is called every so often.

(define (force-output-ports!)
  (let loop ((next (cdr *flush-these-ports*))
             (last *flush-these-ports*))
    (if (not (null? next))
        (let ((port (weak-pointer-ref (car next))))
          (cond ((and port (open-output-port? port))
                 (force-output port)
                 (loop (cdr next) next))
                (else
                 (set-cdr! last (cdr next))
                 (loop (cdr next) last)))))))

;----------------------------------------------------------------

(define buffer-size 1024)

(define (maybe-open-input-file filename)
  (let ((channel (open-channel filename
			       (enum open-channel-option input-file))))
    (if channel
	(input-channel->port channel buffer-size)
	#f)))

(define (maybe-open-output-file filename)
  (let ((channel (open-channel filename
			       (enum open-channel-option output-file))))
    (if channel
	(output-channel->port channel buffer-size)
	#f)))

(define (open-input-file string)
  (or (maybe-open-input-file string)
      (error "can't open for input" string)))

(define (open-output-file string)
  (or (maybe-open-output-file string)
      (error "can't open for output" string)))

